package ac.essex.ooechs.facedetection.singlestage.nodes;

import ac.essex.gp.tree.Node;
import ac.essex.gp.params.NodeConstraints;
import ac.essex.gp.problems.DataStack;
import ac.essex.ooechs.imaging.commons.fast.IntegralImage;
import ac.essex.ooechs.imaging.commons.fast.HaarlikeFeatures;
import ac.essex.ooechs.facedetection.util.AbstractDetectionProblem;

/**
 * Computes the difference between two adjacent rectangles
 *
 * @author Olly Oechsle, University of Essex, Date: 12-Jun-2008
 * @version 1.0
 */
public class ThreeHorizontalRectangleFeature extends Node {

   public ThreeHorizontalRectangleFeature() {
        super(4);
    }

   /**
     * Tells what return type each of the node's children should have
     */
    public int getChildType(int index) {
        return NodeConstraints.PARAMETER;
    }

    public int getReturnType() {
        return NodeConstraints.NUMBER;
    }

    private double px, py, pw, ph;

    public double execute(DataStack data) {

        IntegralImage image = data.getIntegralImage();

        px = child[0].execute(data);
        py = child[1].execute(data);
        pw = child[2].execute(data);
        ph = child[3].execute(data);

        int x = FeatureUtils.getX(px, image.getWidth());
        int y = FeatureUtils.getY(py, image.getHeight());
        int width = FeatureUtils.getWidth(pw, image.getWidth());
        int height = FeatureUtils.getHeight(ph, image.getHeight());

        data.usesImaging = true;
        data.value = image.getHaarlikeFeatures().getThreeRectangleFeature(x,y,width,height, HaarlikeFeatures.HORIZONTALLY_ADJACENT);
        return debugger.record(data.value);

    }

    public String toJava() {
        return "image.getThreeRectangleFeature(x + FeatureUtils.getX(" + px + ", windowWidth), y + FeatureUtils.getY(" + py + ", windowHeight), FeatureUtils.getWidth(" + pw + ", windowWidth), FeatureUtils.getHeight(" + ph + ", windowHeight), HaarRegions.HORIZONTALLY_ADJACENT)";
    }

}
